# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 08:56 (EST)
# Function: the service class
#
# $Id: Service.pm,v 1.25 2002/10/30 19:27:51 jaw Exp $


use TCP;
use UDP;
use Prog;
use Ping;

package Service;
@ISA = qw(MonEl BaseIO);

# to put fresh wood on the fire, chop fuel, carve, cook, pour out wine,
# and do all those services that poor men have to do for their betters.
#   -- Homer, Odyssey

$PHASE_QUANTA = 5;
@probes;	# this is not a useless use of a variable. it is documentation.



$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [@ISA],

    methods => {
	start   => {},
	done    => {},
	timeout => {},
	isdown  => {},
	isup    => {},
	update  => {},
    },
    fields => {
      srvc::hostname => {
	  descr => 'hostname or IP address to test',
	  attrs => ['config', 'inherit'],
      },
      srvc::frequency => {
	  descr => 'how often the test should occur - in seconds',
	  attrs => ['config', 'inherit'],
	  default => 60,
      },
      srvc::phi     => {
	  descr => 'phase - a scheduling parameter',
      },
      srvc::retries => {
	  descr => 'number of times to retry a service before declaring it down',
	  attrs => ['config', 'inherit'],
	  default => 2,
      },
      srvc::retrydelay => {
	  # NB: can be 0 to retry immediately
	  #     default (undef) - wait until next scheduled test
	  descr => 'how soon to retry after a failure, instead of waiting until the next scheduled test',
	  attrs => ['config', 'inherit'],
      },
      srvc::timeout => {
	  descr => 'how long to wait for a response before giving up',
	  attrs => ['config', 'inherit'],
	  default => 60,
      },
#     srvc::status  => {},
      srvc::lasttesttime => {
	  descr => 'time of last test',
      },
      srvc::nexttesttime => {
	  descr => 'time of next test'
      },
      srvc::state => {
	  descr => 'state of the test underway',
      },
      srvc::tries => {
	  descr => 'number of times tried',
      },
      srvc::reason => {
	  descr => 'reason the test failed',
      },
      srvc::result => {
	  descr => 'any result data returned by the test',
      },
      srvc::dead => {},
      srvc::elapsed => {
	  descr => 'how long did the test take', # with very coarse resolution
      },
    },
};

sub new {
    my $class = shift || 'Service';
    my $x = {};
    bless $x, $class;
    $x;
}

sub config {
    my $me = shift;
    my $cf = shift;
    my( $k, $kk, $v, $p, $best );

    $me->init_from_config( $doc, 'srvc' );
    
    $me->{srvc}{phi} = int(rand($me->{srvc}{frequency}/$PHASE_QUANTA)) * $PHASE_QUANTA
	+ int( $PHASE_QUANTA / 2 );

    # What services canst thou do?
    #   -- Shakespesare, King Lear
    # probe, find best match
    foreach $p (@probes){
	my $c = $p->( $me->{name} );
	if( $c ){
	    # probe returns undef or [match, \&config]
	    $best = $c if( !$best || $c->[0] > $best->[0] );
	}
    }

    # service specification errors should only abort this service, trap any errors
    eval {
	# config from best matching probe
	if( $best ){
	    $best->[1]->($me, $cf);
	}else{
	    $cf->error( "unknown service '$me->{name}'" );
	}
    };
    return undef if $@;
    
    $me->{uname} = "$me->{name}_$me->{srvc}{hostname}" unless $me->{uname};
    return undef unless $me->init($cf);

    # the default is different for Service, set if not specified in config
    $me->{notify}{sendnotify} = 1 unless defined $me->{notify}{sendnotify};
    unless( $me->{notify}{sendnotify2} ){
	$me->{notify}{sendnotify} = 0;
    }
    
    $me->reschedule();
    $me;
}

sub reschedule {
    my $me = shift;
    my( $f, $p, $when );

    # My fellow-scholars, and to keep those statutes
    # That are recorded in this schedule here:
    #   -- Shakespeare, Loves Labors Lost
    if( $me->{srvc}{tries} && defined($me->{srvc}{retrydelay}) ){
	$when = $^T + $me->{srvc}{retrydelay};
    }else{
	$f = $me->{srvc}{frequency};
	$p = $me->{srvc}{phi};
    	$when = int( $^T / $f ) * $f + $f + $p;
    }
    $me->{srvc}{nexttesttime} = $when;
    $me->add_timed_func( time => $when,
			 text => 'service start',
			 #func => sub { $me->start() },
			 func => \&me_start,
			 );
}

sub me_start {
    my $me = shift;
    $me->start();
}

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    foreach $k (sort keys %{$me->{srvc}}){
	$v = $me->{srvc}{$k};
	next unless defined $v;
	if( ref($v) ){
	    $v = "#<REF>";
	}else{
	    $v = ::xxx_encode( $v );
	}
	$ctl->write( "srvc::$k:\t$v\n" );
    }
}

sub start {
    my $me = shift;

    # if( $me->{srvc}{dead} ){
    # 	  trace();
    # }
    $me->{srvc}{state} = "starting";
    $me->{srvc}{lasttesttime} = $^T;
    $me->{srvc}{tries} = 1 unless $me->{srvc}{tries};
    $me->debug( "Service start" );
}

sub done {
    my $me = shift;

    $me->shutdown();
    $me->{srvc}{state} = 'done';
    $me->reschedule() unless $me->{srvc}{dead};
}

sub recycle {
    my $me = shift;

    $me->{srvc}{dead} = 1;
    $me->MonEl::recycle();
    $me->BaseIO::recycle();
}

sub isdown {
    my $me = shift;
    my $reason = shift;

    $me->{srvc}{elapsed} = 0;
    $me->{srvc}{reason}  = $reason;
    $me->debug( "Service DOWN - $reason" ) if $reason;
    if( ($me->{srvc}{tries} || 0) <= $me->{srvc}{retries} ){
	# Sometimes to do me service: nine or ten times
	#   -- Shakespeare, Othello
	$me->debug( "Service - retrying" );
	$me->{srvc}{tries} ++;
    }else{
	# For some displeasing service I have done
	#   -- Shakespeare, King Henry IV
	$me->debug( "Service DOWN" );
	$me->{srvc}{tries} = 0;
	$me->update( "down" );
    }
    $me->done();
}

sub isup {
    my $me = shift;

    $me->debug( "Service - UP" );
    $me->{srvc}{elapsed} = $^T - $me->{srvc}{lasttesttime};
    $me->{srvc}{reason} = '';
    $me->{srvc}{tries} = 0;
    $me->update( "up" );
    $me->done();
}

sub update {
    my $me = shift;
    my $st = shift;

    if( $st ne $me->{status} ){
	$me->transition( $st );
    }
}

# override MonEl::transition and jiggle
sub transition {
    my $me = shift;
    my $st = shift;

    if( $st eq 'down' && $me->{watch} && $me->{watch}{watching} ){
	$me->tkt_watch_down();
	return;
    }
    
    if( $st ){
	$me->{prevstatus}   = $me->{status};
	$me->{prevovstatus} = $me->{ovstatus};
	$me->{status}    = $st;
	$me->{ovstatus}  = $st;
	$me->{transtime} = $^T;
	# this may get cleared later if there is an override
	$me->{alarm}     = (($me->{status} eq "down") && $me->{siren}) ? 1 : 0;
	$me->transition2();
	$me->loggit( msg => $me->{status},
		     tag => 'TRANSITION',
		     slp => 1 ) if $me->{prevovstatus} ne $me->{ovstatus};
    }else{
	$me->{ovstatus} = $me->{status};
    }

    $me->transition2();
    $me->transition_propagate();
}

sub jiggle {
    my $me = shift;

    $me->transition_propagate();
}

sub webpage {
    my $me = shift;
    my $fh = shift;
    my $topp = shift;  # NOT USED
    my( $k, $v, $x, $kk );

    print $fh "<!-- start of Service::webpage -->\n";
    
    # object data
    print $fh "<TABLE>\n";
    foreach $k (qw(name ovstatus flags info note comment annotation details)){
	$v = $me->{$k};
	$kk = $k;
	$x = '';
	if( $k eq 'ovstatus' ){
	    $x = " BGCOLOR=\"" . MonEl::web_color($v) . "\"";
	    $kk = 'status';
	}
	print $fh "<TR><TD>$kk</TD><TD$x>$v</TD></TR>\n" if defined($v);
    }
    $me->webpage_more($fh) if $me->can('webpage_more');
    print $fh "</TABLE>\n";

    $me->web_override($fh);
    
    print $fh "<!-- end of Service::webpage -->\n";
}

sub web_page_row_base {
    my $me = shift;
    my $fh = shift;
    my $label = shift;
    
    print $fh "<TR><TD><A HREF=\"", $me->url('func=page'), "\">",
        ($label||$me->{label_left}||$me->{name}), "</A></TD>";

    foreach my $s ('up', 'down', 'override'){
	if( $me->{ovstatus} eq $s ){
	    print $fh "<TD BGCOLOR=\"", MonEl::web_color($s), "\">";
	    print $fh ($me->{label_right}||$me->{name});
	}else{
	    print $fh "<TD>";
	}
	print $fh "</TD>";
    }
    print $fh "</TR>\n";
}

    
################################################################

sub readconfig {
    my $cf  = shift;
    my $mom = shift;
    my $me  = new;
    my( $line, $name, $x, @kinder );
    
    $me->{parents} = [ $mom ] if $mom;

    $line = $cf->nextline();
    ($name) = $line =~ /^\s*[^:\s]+:?\s+([^\{\s]+)/;
    $me->{type} = 'Service';
    $me->{name} = $name;
    $me->{definedinfile} = $cf->{file};
    $me->{definedonline} = $cf->{line};
    
    if( $line =~ /\{/ ){
	while( defined($_ = $cf->nextline()) ){
	    print STDERR "read service: $_\n" if $::opt_d;
	    if( /^\s*\}/ ){
		last;
	    }
	    elsif( /:/ ){
		my($k, $v) = split /:\s*/, $_, 2;
		$cf->warning( "redefinition of parameter '$k'" )
		    if defined $me->{config}{$k};
		$me->{config}{$k} = $v;
	    }
	    else{
		return $cf->error( "invalid entry in config file: '$_'" );
	    }
	}
    }

    return $me->config($cf);
}

sub gen_conf {
    my $me = shift;
    my $in = shift;
    my( $r, $k );

    $r = "Service $me->{name}";
    if( keys %{$me->{config}} ){
	$r .= " {\n";
	foreach $k (sort keys %{$me->{config}}){
	    $r .= "\t$k:\t$me->{config}{$k}";
	    $r .= "\t# unused parameter - typo?"
		unless $me->{confck}{$k};
	    $r .= "\n";
	}
	$r .= "}";
    }
    
    $r .= "\n";
    $r;
}

################################################################
Doc::register( $doc );

1;

