# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 18:25 (EST)
# Function: testing of UDP services
#
# $Id: UDP.pm,v 1.28 2002/10/30 19:27:51 jaw Exp $


package UDP;
@ISA = qw(Service);

use Fcntl;
use Socket;
use SNMP;
use Domain;
use POSIX qw(:errno_h);

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [qw(Service MonEl BaseIO)],
    methods => {},
    fields => {
      udp::port => {
	  descr => 'UDP port to test',
	  attrs => ['config'],
      },
      udp::addr => {
	  descr => 'encoded IP address',
      },
      udp::send => {
	  descr => 'text to send once connected',
	  attrs => ['config'],
      },
      udp::expect => {
	  descr => 'regex to expect',
	  attrs => ['config'],
      },
      udp::rbuffer => {},
      udp::wbuffer => {},
	
    },
};

%config =
(
 # not supported by all radius servers
 RADIUS => {
     port => 1645, send => pack( "CCSx20", 12, $$, 20 ),
     timeout => 10,
 },
 
 DNS => { # RFC 1035, op=status
     port => 53,   send => pack("n CC nnnn", $$, 0x10,0, 0,0,0,0),
     timeout => 12,
 },

 DNSQ => { # djb does not respond to status queries
     port => 53, timeout => 12,
     send => pack( "n CC nnnn  Cnn", $$, 1,0, 1,0,0,0, 0,255,1),
     # op=query, recurse, 1 query, . ANY IN
 },
 
 # test server authority for zone
 Domain => {
     port => 53, timeout => 12,
 },
 
 SNMP => {
     port => 161,  timeout => 5, 
 },

 NTP => { # RFC 2030
     port => 123,  timeout => 10,
     send => pack("CCCC x64", 0x23,0,0,0),
 },

 # NFS = RFC 1094; RPC = RFC 1057; XDR = RFC 1014
 # NFSv2 - NFSPROC_NULL
 NFS => {
     port => 2049, timeout => 10,
     send => pack( "NN NN NN x16", $$, 0,  2, 100003, 2, 0),
     # xid, type, rpcver, prog, ver, func, cred(flavor, len, null), verf(flavor, len, null)
 },

 NFSv3 => {
     port => 2049, timeout => 10,
     send => pack( "NN NN NN x16", $$, 0,  2, 100003, 3, 0),
 },

 );

my $PROTO_UDP = getprotobyname('udp');

sub probe {
    my $name = shift;

    return ( $name =~ /^UDP/ ) ? [ 3, \&config ] : undef;
}

sub config {
    my $me = shift;
    my $cf = shift;
    my( $name, $base );

    bless $me if( ref($me) eq 'Service' );

    $name = $me->{name};
    $name =~ s/^UDP\/?//;
    $base = $name;
    $base =~ s/\/.*//;

    $me->{srvc}{timeout}     = $config{$base}{timeout};
    
    $me->init_from_config( $doc, 'udp' );

    $me->{udp}{port}        ||= $config{$base}{port};
    $me->{udp}{send}        ||= $config{$base}{send};
    $me->{udp}{expect}      ||= $config{$base}{expect};
    
    $me->{udp}{hostname}    = $me->{srvc}{hostname};
    
    if( $name ){
	$me->{uname} = "${name}_$me->{srvc}{hostname}";
    }else{
	$me->{uname} = "UDP_$me->{udp}{port}_$me->{srvc}{hostname}";
    }

    $me->{udp}{addr} = ::resolve( $me->{srvc}{hostname} );
    unless( $me->{udp}{addr} ){
	$me->{transient} = 1;
	return $cf->error( "cannot resolve host $me->{srvc}{hostname}" );
    }

    $me->{label_right_maybe} ||= $name;

    unless( $me->{udp}{port} ){
	$me->{transient} = 1;
	return $cf->error( "Incomplete specification or unknown protocol for Service $name" );
    }

    $me;
}

sub start {
    my $me = shift;
    my( $fh, $i );

    $me->{fd} = $fh = BaseIO::anon_fh();
    $i = socket($fh, PF_INET, SOCK_DGRAM, $PROTO_UDP);
    unless($i){
	my $m = "socket failed: $!";
	::sysproblem( "UDP $m" );
	$me->debug( $m );
	return $me->done();
    }

    $me->baseio_init();
    $me->Service::start();
    
    $me->debug( "UDP start" );
    $i = bind($fh, sockaddr_in(0, INADDR_ANY) );
    unless($i){
	my $m = "bind failed: $!";
	::sysproblem( "UDP $m" );
	$me->debug( $m );
	return $me->done();
    }

    # And they made the plate of the holy crown of pure gold, and wrote upon it a writing
    #   -- exodus 39:30
    $i = send( $fh, $me->{udp}{send}, 0, sockaddr_in($me->{udp}{port}, $me->{udp}{addr}) );
    unless( $i ){
	# QQQ, should I try to sort out errors, down on some, sysproblem on others?
	$me->isdown( "UDP send failed: $!" );
	return;
    }

    $me->debug( "UDP sent data" );

    $me->wantread(1);
    $me->wantwrit(0);
    $me->settimeout( $me->{srvc}{timeout} );
    $me->{srvc}{state} = "waiting";

}

sub readable {
    my $me = shift;
    my( $fh, $i, $l, $e );

    $fh = $me->{fd};
    $i = recv($fh, $l, 8192, 0);
    return $me->isdown( "UDP recv failed: $!" )
	unless defined($i);

    $me->debug( "UDP recv data" );
    $me->{udp}{rbuffer} = $l;		# for debugging
    
    # yet I will read the writing unto the king, and make known to him the interpretation
    #   -- daniel 5:17

    $e = $me->{udp}{expect};
    if( $e ){
	if( $l =~ /$e/ ){
	    $me->isup();
	}else{
	    # O, when she had writ it and was reading it over, she
	    # found Benedick and Beatrice between the sheet?
	    #   -- Shakespeare, Much Ado About Nothing
	    $me->isdown( "UDP - did not get expected data" );
	}
    }else{
	$me->isup();
    }
}

sub timeout {
    my $me = shift;

    $me->isdown( "UDP timeout" );
}

################################################################
# and also object methods
################################################################

# sub done {
#     my $me = shift;
# 
#     $me->Service::done();
# }
# 
# sub isdown {
#     my $me = shift;
#     my $reason = shift;
# 
#     $me->Service::isdown($reason);
# }
# 
# sub isup {
#     my $me = shift;
# 
#     $me->Service::isup();
# }

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->Service::about_more($ctl);

    foreach $b ('udp'){
	next unless $me->{$b};
	foreach $k (sort keys %{$me->{$b}}){
	    $v = $me->{$b}{$k};
	    next unless defined $v;
	    if( ref($v) ){
		$v = "#<REF>";
	    }else{
		# QQQ - or some other encoding scheme?
		$v = ::xxx_encode( $v );
	    }
	    $ctl->write( "${b}::$k:\t$v\n" );
	}
    }
}

sub webpage_more {
    my $me = shift;
    my $fh = shift;
    my( $k, $v );
    
    foreach $k (qw(hostname port)){
	$v = $me->{udp}{$k};
	print $fh "<TR><TD>$k</TD><TD>$v</TD></TR>\n" if defined($v);
    }
}

################################################################
# global config
################################################################
Doc::register( $doc );
push @Service::probes, \&probe;

1;

